#include "cpu.h"
#include "pic32_c.h"
#include "init.h"



/* Basic memory and pin initialisation. */

void init_memory(void)
{
    /*
    Configure RAM.
    See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
    */

    uint32_t config = REG(BMXCON);

    /* Set zero wait states for address setup. */

    config &= ~(1 << 6);        /* BMXCON<6> = BMXWSDRM = 0 */

    /* Set bus arbitration mode. */

    config &= ~0b111;
    config |= 0b010;            /* BMXCON<2:0> = BMXARB<2:0> = 2 */

    REG(BMXCON) = config;
}

void init_pins(void)
{
    /* DEVCFG0<2> also needs setting to 0 before the program is run. */

    CLR_REG(CFGCON, 1 << 3);    /* CFGCON<3> = JTAGEN = 0 */
}

void init_outputs(void)
{
    /* Remove analogue features from pins. */

    REG(ANSELA) = 0;
    REG(ANSELB) = 0;

    /* Set pins as outputs. */

    REG(TRISA) = 0;
    REG(TRISB) = 0;

    /* Clear outputs. */

    REG(PORTA) = 0;
    REG(PORTB) = 0;
}



/* Peripheral pin configuration. */

void config_uart(void)
{
    /* NOTE: Configuring UART1 for specific pins. */

    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}

void lock_config(void)
{
    SET_REG(CFGCON, 1 << 13);   /* IOLOCK = 1 */

    /* Lock the configuration again. */

    REG(SYSKEY) = 0x33333333;
}

void unlock_config(void)
{
    /* Unlock the configuration register bits. */

    REG(SYSKEY) = 0;
    REG(SYSKEY) = 0xAA996655;
    REG(SYSKEY) = 0x556699AA;

    CLR_REG(CFGCON, 1 << 13);   /* IOLOCK = 0 */
}



/* Convenience operations. */

void interrupts_on(void)
{
    init_interrupts();
    enable_interrupts();
    handle_error_level();
}



/* DMA configuration. */

void init_dma(void)
{
    /* Disable DMA interrupts. */

    CLR_REG(DMAIEC, 0b1111 << DMAINTBASE); /* DMA3IE...DMA0IE = 0 */

    /* Clear DMA interrupt flags. */

    CLR_REG(DMAIFS, 0b1111 << DMAINTBASE); /* DMA3IF...DMA0IF = 0 */

    /* Enable DMA. */

    SET_REG(DMACON, 1 << 15);
}

/* Initialise the given channel. */

void dma_init(int channel, int auto_enable, uint8_t pri)
{
    if (channel > DCHMAX)
        return;

    /* Initialise a channel. */

    REG(DMA_REG(channel, DCHxCON)) = (auto_enable ? (1 << 4) : 0) | (pri & 0b11);
    REG(DMA_REG(channel, DCHxECON)) = 0;
    REG(DMA_REG(channel, DCHxINT)) = 0;
}

/* Configure a channel's initiation interrupt. */

void dma_set_interrupt(int channel, uint8_t int_num, int enable)
{
    if (channel > DCHMAX)
        return;

    /* Allow an interrupt to trigger the transfer. */

    REG(DMA_REG(channel, DCHxECON)) = (int_num << 8) |
                                      ((enable ? 1 : 0) << 4);
}

/* Set a channel's transfer parameters. */

void dma_set_transfer(int channel,
                  uint32_t source_start_address, uint16_t source_size,
                  uint32_t destination_start_address, uint16_t destination_size,
                  uint16_t cell_size)
{
    if (channel > DCHMAX)
        return;

    REG(DMA_REG(channel, DCHxSSIZ)) = source_size;
    REG(DMA_REG(channel, DCHxSSA)) = source_start_address;
    REG(DMA_REG(channel, DCHxDSIZ)) = destination_size;
    REG(DMA_REG(channel, DCHxDSA)) = destination_start_address;
    REG(DMA_REG(channel, DCHxCSIZ)) = cell_size;
}

/* Configure interrupts caused by the channel. */

void dma_init_interrupt(int channel, uint8_t conditions,
                        uint8_t int_pri, uint8_t int_sub)
{
    if (channel > DCHMAX)
        return;

    /* Produce an interrupt for the provided conditions. */

    REG(DMA_REG(channel, DCHxINT)) = conditions << 16;

    /* Set interrupt priorities. */

    REG(DMAIPC) = (REG(DMAIPC) & ~(DMA_IPC_PRI(channel, 7, 3))) |
                  DMA_IPC_PRI(channel, int_pri, int_sub);

    /* Enable interrupt. */

    SET_REG(DMAIEC, 1 << (channel + DMAINTBASE));
}

/* Enable a DMA channel. */

void dma_on(int channel)
{
    if (channel > DCHMAX)
        return;

    /* Enable channel. */

    SET_REG(DMA_REG(channel, DCHxCON), 1 << 7);
}

/* UART configuration. */

void uart_init(int channel, uint32_t baudrate)
{
    /* NOTE: Configured in the initial payload. */

    uint32_t FPB = 24000000;

    if ((channel < UARTMIN) || (channel > UARTMAX))
        return;

    /* Disable the UART (ON). */

    CLR_REG(UART_REG(channel, UxMODE), 1 << 15);

    /* Set the baud rate. For example:

       UxBRG<15:0> = BRG
                   = (FPB / (16 * baudrate)) - 1
                   = (24000000 / (16 * 115200)) - 1
                   = 12
    */

    REG(UART_REG(channel, UxBRG)) = (FPB / (16 * baudrate)) - 1;

    /* Disable interrupt (UxRIE) and clear flag (UxRIF). */

    CLR_REG(UARTIEC, UART_INT_FLAGS(channel, UxRIE));
    CLR_REG(UARTIFS, UART_INT_FLAGS(channel, UxRIF));
}

/* Configure interrupts caused by the UART. */

void uart_init_interrupt(int channel, uint8_t pri, uint8_t sub)
{
    if ((channel < UARTMIN) || (channel > UARTMAX))
        return;

    /* Set priorities: UxIP = pri; UxIS = sub */

    REG(UART_IPC_REG(channel)) = (REG(UART_IPC_REG(channel)) & ~UART_IPC_PRI(channel, 7, 3)) |
                                 UART_IPC_PRI(channel, pri, sub);

    /* Enable interrupt (UxRIE). */

    SET_REG(UARTIEC, UART_INT_FLAGS(channel, UxRIE));
}

/* Enable a UART. */

void uart_on(int channel)
{
    if ((channel < UARTMIN) || (channel > UARTMAX))
        return;

    /* Enable receive (URXEN) and transmit (UTXEN). */

    SET_REG(UART_REG(channel, UxSTA), (1 << 12) | (1 << 10));

    /* Start UART. */

    SET_REG(UART_REG(channel, UxMODE), 1 << 15);
}



/* Utility functions. */

/* Return encoded interrupt priorities. */

static uint8_t PRI(uint8_t pri, uint8_t sub)
{
    return ((pri & 0b111) << 2) | (sub & 0b11);
}

/* Return encoded DMA interrupt priorities for combining with a register. */

uint32_t DMA_IPC_PRI(int channel, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (DCHIPCBASE + (channel - DCHMIN) * DCHIPCSTEP);
}

/* Return encoded UART interrupt priorities for combining with a register. */

uint32_t UART_IPC_PRI(int channel, uint8_t pri, uint8_t sub)
{
    return PRI(pri, sub) << (channel == 1 ? UART1IPCBASE : UART2IPCBASE);
}

/* Return the UART interrupt priorities register. */

uint32_t UART_IPC_REG(int channel)
{
    return channel == 1 ? UART1IPC : UART2IPC;
}

/* Return the UART interrupt flags shift offset. */

int UART_INT_FLAGS(int channel, uint8_t flags)
{
    return (flags & 0b111) << (UARTINTBASE + (channel - UARTMIN) * UARTINTSTEP);
}
